;---------------------------------------------------------------------------;
; Extended itoa() and puts() function                    (C)ChaN, 2005

.nolist
#include <avr/io.h>	// Include device specific definitions.
.list

#ifdef SPM_PAGESIZE	// The device has "lpm Rd,Z+" and "movw".
.macro	_LPMI	reg
	lpm	\reg, Z+
.endm
.macro	_MOVW	dh,dl, sh,sl
	movw	\dl, \sl
.endm
#else			// The device does not have "lpm Rd,Z+" nor "movw".
.macro	_LPMI	reg
	lpm
	mov	\reg, r0
	adiw	ZL, 1
.endm
.macro	_MOVW	dh,dl, sh,sl
	mov	\dl, \sl
	mov	\dh, \sh
.endm
#endif


;---------------------------------------------------------------------------
; Stub function to forward to user output function
;
;Prototype: void xputc (char chr);  // a character to be output
;Size: 5/5 words

.section .bss

.global xfunc_out	; xfunc_out must be initialized before using this module.
xfunc_out:	.ds.w	1

.section .text


.global xputc
.func xputc
xputc:
	lds	ZL, xfunc_out+0	;Pointer to the registered output function.
	lds	ZH, xfunc_out+1	;/
	ijmp
.endfunc


;---------------------------------------------------------------------------
; Extended direct numeral string output
;
;Prototype: void xitoa (int value,          // value to be output
;                       char base,          // base N
;                       char digits );      // number of digits
;Size: 53/53 words
;
; value  base  digits   output
;   100    10       6   "   100"
;   100    10      -6   "000100"
;   100    10       0   "100"
; 65535    10       0   "65535"
; 65535   -10       0   "-1"
;  1024    16       0   "400"
;  1024    16      -4   "0400"
;  0x55     2      -8   "01010101"

.global xitoa
.func xitoa
xitoa:
				;r25:r24 = value, r22 = base, r20 = digits
	clr	r21		;r21 = stack level
	ldi	r30, ' '	;r30 = sign
	ldi	r31, ' '	;r31 = filler
	sbrs	r22, 7		;When base indicates signd format and the value
	rjmp	0f		;is minus, add a '-'.
	neg	r22		;
	sbrs	r25, 7		;
	rjmp	0f		;
	ldi	r30, '-'	;
	com	r24		;
	com	r25		;
	adc	r24, r1		;
	adc	r25, r1		;/
0:	sbrs	r20, 7		;When digits indicates zero filled,
	rjmp	1f		;filler is '0'.
	neg	r20		;
	ldi	r31, '0'	;/
				;----- string conversion loop
1:	ldi	r23, 16		;r26 = r25:r24 % r22
	clr	r26		;r25:r24 /= r22
2:	lsl	r24		;
	rol	r25		;
	rol	r26		;
	cp	r26, r22	;
	brcs	3f		;
	sub	r26, r22	;
	inc	r24		;
3:	dec	r23		;
	brne	2b		;/
	cpi	r26, 10		;r26 is a numeral digit '0'-'F'
	brcs	4f		;
	subi	r26, -7		;
4:	subi	r26, -'0'	;/
	push	r26		;Stack it
	inc	r21		;/
	cp	r24, r1		;Repeat until r25:r24 gets zero
	cpc	r25, r1		;
	brne	1b		;/

	cpi	r30, '-'	;Minus sign if needed
	brne	5f		;
	push	r30		;
	inc	r21		;/
5:	cp	r21, r20	;Filler
	brcc	6f		;
	push	r31		;
	inc	r21		;
	rjmp	5b		;/

6:	pop	r24		;Flush stacked digits and exit
	push	r21		;
	rcall	xputc		;
	pop	r21		;
	dec	r21		;
	brne	6b		;/

	ret
.endfunc


;---------------------------------------------------------------------------
; Direct ROM string output
;
;Prototype: void xputs (const prog_char *str // rom string to be output
;			);
;Size: 14/11 words

.global xputs
.func xputs
xputs:
	_MOVW	ZH,ZL, r25,r24	; Z = pointer to rom string

1:	_LPMI	r24
	cpi	r24, 0
	breq	2f
	push	ZH
	push	ZL
	rcall	xputc
	pop	ZL
	pop	ZH
	rjmp	1b
2:	ret
.endfunc


/*
;---------------------------------------------------------------------------;
; Formatted string output
;
;Prototype: void xprintf (const prog_char *format, ...);
;Size: 74/67 words

.global xprintf
.func xprintf
xprintf:
	push	YH
	push	YL
	in	YL, _SFR_IO_ADDR(SPL)
#ifdef SPH
	in	YH, _SFR_IO_ADDR(SPH)
#else
	clr	YH
#endif
	adiw	YL, 5		;Y = pointer to arguments
	ld	ZL, Y+		;Z = pointer to format string
	ld	ZH, Y+		;/

0:	_LPMI	r24		;Get a format char
	cpi	r24, 0		;End of format string?
	breq	90f		;/
	cpi	r24, '%'	;Is format?
	breq	20f		;/
1:	push	ZH		;Put a normal character
	push	ZL		;
	rcall	xputc		;
10:	pop	ZL		;
	pop	ZH		;
	rjmp	0b		;/
90:	pop	YL
	pop	YH
	ret

20:	ldi	r20, 0		;r20: digits
	clt			;T: filler
21:	_LPMI	r24		;Get flags
	cpi	r24, '%'	;Is '%'?
	breq	1b		;/
	cpi	r24, '0'	;Zero filled?
	brne	23f		;
	set			;/
22:	_LPMI	r24		;Get width
23:	cpi	r24, '9'+1	;
	brcc	24f		;
	subi	r24, '0'	;
	brcs	90b		;
	lsl	r20		;
	mov	r0, r20		;
	lsl	r20		;
	lsl	r20		;
	add	r20, r0		;
	add	r20, r24	;
	rjmp	22b		;/

24:	mov	r23, r24	;r23 = type
	ld	r24, Y+		;r25:r24 = value
	ld	r25, Y+		;/
	cpi	r23, 'c'	;Is type character?
	breq	1b		;/
	push	ZH
	push	ZL
	cpi	r23, 's'	;Is type string?
	breq	50f		;/
	cpi	r23, 'X'	;Is type hexdecimal?
	ldi	r22, 16		;
	breq	40f		;/
	cpi	r23, 'u'	;Is type unsigned decimal?
	ldi	r22, 10		;
	breq	40f		;/
	cpi	r23, 'd'	;Is type signed decimal?
	ldi	r22, -10	;
	breq	40f		;/
	cpi	r23, 'b'	;Is type binary?, or ignore.
	ldi	r22, 2		;
	brne	10b		;/
40:	brtc	41f		;Output the value
	neg	r20		;
41:	rcall	xitoa		;
	rjmp	10b		;/
50:	rcall	xputs		;Output the ROM string
	rjmp	10b		;/
.endfunc
*/

